"""Handlers to pull in datasets and perform preprocessing."""

# pylint: disable=consider-using-with, too-many-return-statements

import logging
import os
from zipfile import BadZipFile, ZipFile

import numpy as np
import pandas as pd
import pylab as plt
from sklearn.datasets import fetch_openml, load_iris
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

logging.basicConfig(level="DEBUG")

logger = logging.getLogger(__file__)


PROJECT_ROOT_FOLDER = os.path.dirname(os.path.dirname(__file__))
logger.info("ROOT PROJECT FOLDER = %s", PROJECT_ROOT_FOLDER)


class UnknownDataset(Exception):
    """Exception raised if the user passes a name that we don't recognise."""


class DataNotAvailable(Exception):
    """Exception raised if the user asks for a dataset that they do not have."""


def get_data_sklearn(  # pylint: disable = too-many-branches
    dataset_name: str, data_folder: str = os.path.join(PROJECT_ROOT_FOLDER, "data")
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get data in a format sensible for sklearn.

    User passes a name and that dataset is returned as a tuple of pandas
    DataFrames (data, labels).

    Parameters
    ----------
    dataset_name : str
        The name of the dataset to load
    data_folder : str
        The name of the local folder in which data is stored.

    Returns
    -------
    X : pd.DataFrame
        The input dataframe -- rows are examples, columns variables
    y : pd.DataFrame
        the target dataframe -- has a single column containing the target values

    Notes
    -----
    The following datasets are available:
    mimic2-iaccd (requires data download)
    in-hospital-mortality (requires data download)
    medical-mnist-ab-v-br-100 (requires data download)
    medical-mnist-ab-v-br-500 (requires data download)
    medical-mnist-all-100 (requires data download)
    indian liver (requires data download)
    synth-ae (requires data download)
    synth-ae-small (requires data download)
    nursery (downloads automatically)
    iris (available out of the box via sklearn)

    Datasets can be normalised by adding the following prefixes:
    standard: standardises all columns to have zero mean and unit variance.
    minmax: standardises all columns to have values between 0 and 1.
    round: rounds continues features to have 3dp

    These can be nested.

    Examples
    --------
    .. code-block:: python

        # pull the mimic2-iaccd data
        X, y = get_data_sklearn("mimic2-iaccd")

        # pull the iris data and round continuous features
        X, y = get_data_sklearn("minmax iris")
    """
    logger.info("DATASET FOLDER = %s", data_folder)

    if dataset_name.startswith("standard"):
        sub_name = dataset_name.split("standard")[1].strip()
        feature_df, target_df = get_data_sklearn(sub_name, data_folder)
        for column in feature_df.columns:
            col_mean = feature_df[column].mean()
            col_std = np.sqrt(feature_df[column].var())
            feature_df[column] = feature_df[column] - col_mean
            feature_df[column] = feature_df[column] / col_std
        return feature_df, target_df

    if dataset_name.startswith("minmax"):
        sub_name = dataset_name.split("minmax")[1].strip()
        feature_df, target_df = get_data_sklearn(sub_name, data_folder)
        for column in feature_df.columns:
            col_min = feature_df[column].min()
            col_range = feature_df[column].max() - col_min
            feature_df[column] = feature_df[column] - col_min
            feature_df[column] = feature_df[column] / col_range
        return feature_df, target_df

    if dataset_name.startswith("round"):
        sub_name = dataset_name.split("round")[1].strip()
        logger.debug(sub_name)
        feature_df, target_df = get_data_sklearn(sub_name, data_folder)
        column_dtype = feature_df.dtypes

        for i, column in enumerate(feature_df.columns):
            if column_dtype[i] == "float64":
                feature_df[column] = feature_df[column].round(decimals=3)
        return feature_df, target_df

    if dataset_name == "mimic2-iaccd":
        return _mimic_iaccd(data_folder)
    if dataset_name == "in-hospital-mortality":
        return _in_hospital_mortality(data_folder)
    if dataset_name == "medical-mnist-ab-v-br-100":
        return _medical_mnist_loader(data_folder, 100, ["AbdomenCT", "BreastMRI"])
    if dataset_name == "medical-mnist-ab-v-br-500":
        return _medical_mnist_loader(data_folder, 500, ["AbdomenCT", "BreastMRI"])
    if dataset_name == "medical-mnist-all-100":
        return _medical_mnist_loader(
            data_folder,
            100,
            ["AbdomenCT", "BreastMRI", "CXR", "ChestCT", "Hand", "HeadCT"],
        )
    if dataset_name == "indian liver":
        return _indian_liver(data_folder)
    if dataset_name == "synth-ae":
        return _synth_ae(data_folder)
    if dataset_name == "synth-ae-small":
        return _synth_ae(data_folder, 200)
    if dataset_name == "synth-ae-large":
        return _synth_ae(data_folder, 500000)
    if dataset_name == "synth-ae-extra-large":
        return _synth_ae(data_folder, 2000000)
    if dataset_name == "synth-ae-XXL":
        return _synth_ae(data_folder, 50000000)
    if dataset_name == "nursery":
        return _nursery()
    if dataset_name == "iris":
        return _iris()
    if dataset_name == "RDMP":
        return _rdmp(data_folder)
    raise UnknownDataset(dataset_name)


def _iris() -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get the Sklearn iris data - just first two classes."""
    X, y = load_iris(return_X_y=True, as_frame=True)
    X = X[y < 2]
    y = y[y < 2]
    return X, pd.DataFrame(y)


def _nursery() -> tuple[pd.DataFrame, pd.DataFrame]:
    """Return the sklearn nursery dataset."""
    data = fetch_openml(data_id=26, as_frame=True)

    target_encoder = LabelEncoder()
    target_vals = target_encoder.fit_transform(data["target"].values)
    target_dataframe = pd.DataFrame({"target": target_vals})

    feature_encoder = OneHotEncoder()
    x_encoded = feature_encoder.fit_transform(data["data"]).toarray()
    feature_dataframe = pd.DataFrame(
        x_encoded, columns=feature_encoder.get_feature_names_out()
    )

    return feature_dataframe, target_dataframe


def _images_to_ndarray(
    images_dir: str, number_to_load: int, label: int, flatten: bool = True
) -> tuple[np.array, np.array]:
    """Get number_to_load images from the images_dir and create arrays.

    Patched to support non-flattened images.
    Same behaviour as before except if called with flatten=False explicitly.
    """
    folder_path = images_dir + os.sep
    images_names = sorted(os.listdir(folder_path))
    images_names = images_names[:number_to_load]
    # fix f or macosx
    if ".DS_Store" in images_names:
        images_names.remove(".DS_Store")

    if flatten:
        np_images = np.array(
            [plt.imread(folder_path + img).flatten() for img in images_names]
        )
    else:
        np_images = np.array([plt.imread(folder_path + img) for img in images_names])
    labels = np.ones((len(np_images), 1), int) * label
    return (np_images, labels)


def _medical_mnist_loader(  # pylint: disable = too-many-locals
    data_folder: str, n_per_class: int, classes: list[str]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get Medical MNIST into pandas format.

    Borrows heavily from: https://www.kaggle.com/harelshattenstein/medical-mnist-knn
    Creates a binary classification.
    """
    base_folder = os.path.join(
        data_folder,
        "kaggle-medical-mnist",
        "archive",
    )

    zip_file = os.path.join(data_folder, "kaggle-medical-mnist", "archive.zip")

    logger.debug(base_folder, data_folder)
    if not any([os.path.exists(base_folder), os.path.exists(zip_file)]):
        help_message = f"""
Data file {base_folder} does not exist. Please download fhe file from:
https://www.kaggle.com/andrewmvd/medical-mnist
and place it in the correct folder. It unzips the file first.
        """
        raise DataNotAvailable(help_message)

    if os.path.exists(base_folder):
        pass
    elif os.path.exists(zip_file):
        try:
            with ZipFile(zip_file) as zip_handle:
                zip_handle.extractall(base_folder)
                logger.debug("Extracted all")
        except BadZipFile:
            logger.error("Encountered bad zip file")
            raise

    labels_dict = {
        0: "AbdomenCT",
        1: "BreastMRI",
        2: "CXR",
        3: "ChestCT",
        4: "Hand",
        5: "HeadCT",
    }

    reverse_labels_dict = {v: k for k, v in labels_dict.items()}

    for i, class_name in enumerate(classes):
        label = reverse_labels_dict[class_name]
        x_images, y_images = _images_to_ndarray(
            os.path.join(base_folder, class_name), n_per_class, label
        )

        if i == 0:
            all_x = x_images
            all_y = y_images
        else:
            all_x = np.vstack((all_x, x_images))
            all_y = np.vstack((all_y, y_images))

    return (pd.DataFrame(all_x), pd.DataFrame(all_y))


def _synth_ae(
    data_folder: str, n_rows: int = 5000
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get synth ae data.

    First norws (default 5000) rows from the Synthetic A&E data from NHS England
    https://data.england.nhs.uk/dataset/a-e-synthetic-data/resource/81b068e5-6501-4840-a880-a8e7aa56890e # pylint: disable=line-too-long.
    """
    file_path = os.path.join(data_folder, "AE_England_synthetic.csv")

    if not os.path.exists(file_path):
        help_message = f"""
Data file {file_path} does not exist. Please download the file from:
https://data.england.nhs.uk/dataset/a-e-synthetic-data/resource/81b068e5-6501-4840-a880-a8e7aa56890e

Alternatively, download the file directly from the following URL:
https://nhsengland-direct-uploads.s3-eu-west-1.amazonaws.com/A%26E+Synthetic+Data.7z

Unzip it (7z) and then copy the .csv file into your data folder.
    """
        raise DataNotAvailable(help_message)

    input_data = pd.read_csv(file_path, nrows=n_rows)
    columns_to_drop = [
        "AE_Arrive_Date",
        "AE_Arrive_HourOfDay",
        "Admission_Method",
        "ICD10_Chapter_Code",
        "Treatment_Function_Code",
        "Length_Of_Stay_Days",
        "ProvID",
    ]
    input_data.drop(columns_to_drop, axis=1, inplace=True)

    # Remove any rows with NAs in the remaining columns
    input_data.dropna(axis=0, inplace=True)

    # One-hot encode some columns
    encode_columns = ["Age_Band", "AE_HRG"]
    encode_data = input_data[encode_columns]
    input_data.drop(encode_columns, axis=1, inplace=True)

    oh = OneHotEncoder()
    oh.fit(encode_data)
    onehot_df = pd.DataFrame(
        oh.transform(encode_data).toarray(),
        columns=oh.get_feature_names_out(),
        index=input_data.index,
    )

    input_data = pd.concat([input_data, onehot_df], axis=1)

    X = input_data.drop(["Admitted_Flag"], axis=1)
    y = input_data[["Admitted_Flag"]]

    return (X, y)


def _indian_liver(data_folder: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get Indian Liver Patient Dataset.

    https://archive.ics.uci.edu/ml/machine-learning-databases/00225/Indian%20Liver%20Patient%20Dataset%20(ILPD).csv # pylint: disable=line-too-long.
    """
    file_path = os.path.join(data_folder, "Indian Liver Patient Dataset (ILPD).csv")
    if not os.path.exists(file_path):
        help_message = f"""
Data file {file_path} does not exist. Please download fhe file from:
https://archive.ics.uci.edu/ml/datasets/ILPD+(Indian+Liver+Patient+Dataset
and place it in the correct folder.
        """
        raise DataNotAvailable(help_message)

    column_names = [
        "age",
        "gender",
        "total Bilirubin",
        "direct Bilirubin",
        "Alkphos",
        "SGPT",
        "SGOT",
        "total proteins",
        "albumin",
        "A/G ratio",
        "class",
    ]

    liver_data = pd.read_csv(file_path, names=column_names, index_col=False)

    liver_data.gender.replace("Male", 0, inplace=True)
    liver_data.gender.replace("Female", 1, inplace=True)

    liver_data.dropna(axis=0, inplace=True)

    liver_labels = liver_data["class"]
    liver_data.drop(["class"], axis=1, inplace=True)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(liver_labels.values)
    liver_labels = pd.DataFrame({"class": encoded_labels})
    return (liver_data, liver_labels)


def _in_hospital_mortality(data_folder: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get In-hospital mortality data.

    See: https://datadryad.org/stash/dataset/doi:10.5061/dryad.0p2ngf1zd.
    """
    # Check the data has been downloaded.
    # If not, throw an exception with instructions on how to
    # download, and where to store
    files = ["data01.csv", "doi_10.5061_dryad.0p2ngf1zd__v5.zip"]
    file_path = [os.path.join(data_folder, f) for f in files]
    print(file_path)

    if not any(os.path.exists(fp) for fp in file_path):
        help_message = f"""
Data file {file_path[0]} or {file_path[1]} does not exist. Please download the file from:
https://datadryad.org/stash/dataset/doi:10.5061/dryad.0p2ngf1zd
and place it in the correct folder. It works with either the zip file or uncompressed.
Alternatively download the data file from this URL:
        https://datadryad.org/stash/downloads/file_stream/773992
and then change the name of the file 773992 to data01.csv.
        """
        raise DataNotAvailable(help_message)

    if os.path.exists(file_path[1]):
        input_data = pd.read_csv(ZipFile(file_path[1]).open("data01.csv"))
    else:
        input_data = pd.read_csv(file_path[0])
    clean_data = input_data.dropna(axis=0, how="any").drop(columns=["group", "ID"])
    target = "outcome"
    labels = clean_data[target]
    features = clean_data.drop([target], axis=1)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(labels.values)
    labels = pd.DataFrame({"outcome": encoded_labels})

    return (features, labels)


def _mimic_iaccd(data_folder: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get the mimic_iaccd data and perform preprocessing."""
    # Check the data has been downloaded.
    # If not throw an exception with instructions on how to
    # download, and where to store
    file_path = os.path.join(data_folder, "mimic2-iaccd", "1.0", "full_cohort_data.csv")
    print(file_path, os.path.exists(file_path))

    if not os.path.exists(file_path):
        help_message = f"""
        The MIMIC2-iaccd data is not available in {data_folder}.
        The following file should exist: {file_path}.
        Please download from https://physionet.org/files/mimic2-iaccd/1.0/full_cohort_data.csv?download
        and rename the file to full_cohort_data.csv.
        """
        raise DataNotAvailable(help_message)

    # File exists, load and preprocess#
    logger.info("Loading mimic2-iaccd")
    input_data = pd.read_csv(file_path)

    logger.info("Preprocessing")
    # remove columns non-numerical and repetitive or uninformative data for the analysis
    col = [
        "service_unit",
        "day_icu_intime",
        "hosp_exp_flg",
        "icu_exp_flg",
        "day_28_flg",
    ]
    # service_num is the numerical version of service_unit
    # day_icu_intime_num is the numerical version of day_icu_intime
    # the other columns are to do with death and are somewhat repetitive with censor_flg
    input_data.drop(col, axis=1, inplace=True)

    # drop columns with only 1 value
    input_data.drop("sepsis_flg", axis=1, inplace=True)

    # drop NA by row
    input_data.dropna(axis=0, inplace=True)

    # extract target
    target = "censor_flg"
    y = input_data[target]
    X = input_data.drop([target, "mort_day_censored"], axis=1)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(y.values)
    y = pd.DataFrame({"censor_flag": encoded_labels})

    return (X, y)


def _rdmp(  # pylint: disable=too-many-locals, too-many-statements
    data_folder: str,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Get the RDMP dataset."""

    def find_age(row: pd.Series) -> int:
        date_ = pd.to_datetime("01/06/2020")
        if row.date_of_death != row.date_of_death:
            age = np.floor((date_ - row.date_of_birth).days / 365.25)
        else:
            age = np.floor((row.date_of_death - row.date_of_birth).days / 365.25)
        return age

    def hospital_days(row: pd.Series) -> int:
        if row.DischargeDate == row.DischargeDate:
            t = row.DischargeDate - row.AdmissionDate
            days = t.days + round(((t.seconds / 60) / 60) / 24)
        else:
            days = 0
        return days

    processed_data_file = "rdmp_binary.csv"
    os.path.join(data_folder, "RDMP")
    if os.path.exists(os.path.join(data_folder, "RDMP", processed_data_file)):
        logger.info("Loading processed RDMP file.")
        # load data processed csv file
        df = pd.read_csv(os.path.join(data_folder, "RDMP", processed_data_file))
    else:
        logger.info("Processing RDMP synthetic data")
        file_list = [
            "CarotidArteryScan.csv",
            "Demography.csv",
            "HospitalAdmissions.csv",
        ]
        files_path = [os.path.join(data_folder, "RDMP", f) for f in file_list]
        found = [os.path.exists(file_path) for file_path in files_path]
        not_found = [
            file_path for file_path in files_path if not os.path.exists(file_path)
        ]

        if not all(found):
            help_message = f"""
            Some or all data files do not exist. Please download the files from RDMP,
        and place it in the correct folder.

            Missing files are:
            {not_found}
            """
            raise DataNotAvailable(help_message)

        # Load data
        headers0 = [
            "R_CC_STEN_A",
            "R_CC_STEN_B",
            "R_CC_STEN_C",
            "R_CC_STEN_D",
            "R_CC_STEN_S",
            "L_IC_STEN_A",
            "L_IC_STEN_B",
            "L_IC_STEN_C",
            "L_IC_STEN_D",
            "L_IC_STEN_S",
            "R_IC_STEN_A",
            "R_IC_STEN_B",
            "R_IC_STEN_C",
            "R_IC_STEN_D",
            "R_IC_STEN_S",
            "PatientID",
            "L_CC_STEN_S",
            "L_CC_STEN_D",
            "L_CC_STEN_B",
            "L_BD_RATIO",
            "L_AC_RATIO",
            "R_BD_RATIO",
            "R_AC_RATIO",
            "L_CC_STENOSIS",
            "L_CC_PEAK_SYS",
            "L_CC_END_DIA",
            "L_IC_STENOSIS",
            "L_IC_PEAK_SYS",
            "L_IC_END_DIA",
            "L_EC_STENOSIS",
            "L_PLAQUE",
            "L_SYMPTOMS",
            "L_BRUIT",
            "L_CC_STEN_A",
            "ON_STEN_STUDY",
            "R_VERT_ARTERY",
            "R_BRUIT",
            "R_SYMPTOMS",
            "R_PLAQUE",
            "L_CC_STEN_C",
            "R_EC_STENOSIS",
            "R_IC_PEAK_SYS",
            "R_IC_STENOSIS",
            "R_CC_END_DIA",
            "R_CC_PEAK_SYS",
            "R_CC_STENOSIS",
            "L_VERT_ARTERY",
            "R_IC_END_DIA",
        ]
        headers1 = [
            "chi",
            "sex",
            "current_address_L2",
            "date_of_death",
            "date_of_birth",
        ]
        headers2 = [
            "chi",
            "AdmissionDate",
            "DischargeDate",
            "MainCondition",
            "OtherCondition1",
            "OtherCondition2",
            "OtherCondition3",
            "MainOperation",
            "MainOperationB",
            "OtherOperation1",
            "OtherOperation1B",
            "OtherOperation2",
            "OtherOperation2B",
            "OtherOperation3",
            "OtherOperation3B",
        ]

        # Process first file
        df = pd.read_csv(files_path[0], usecols=headers0, encoding="ISO 8859-1")
        # Change name to be the same in all files
        df.rename(columns={"PatientID": "chi"}, inplace=True)
        df = df.groupby(["chi"]).max()

        # Process second file
        df_ = pd.read_csv(files_path[1], usecols=headers1)
        df_["date_of_birth"] = pd.to_datetime(df_["date_of_birth"])
        df_["date_of_death"] = pd.to_datetime(df_["date_of_death"])
        df_ = df_.groupby(["chi"]).max()

        # Merge first and second file
        df = df.merge(df_, how="inner", on="chi", suffixes=(False, False))
        del df_

        # Process third file
        df__ = pd.read_csv(files_path[2], usecols=headers2, encoding="ISO 8859-1")
        df__["AdmissionDate"] = pd.to_datetime(df__["AdmissionDate"])
        df__["DischargeDate"] = pd.to_datetime(df__["DischargeDate"])
        df__["days_in_hospital"] = df__.apply(hospital_days, axis=1)
        number_stays = df__.groupby(["chi"]).count()["AdmissionDate"]
        dih = df__.groupby(["chi"])["days_in_hospital"].sum()
        nc = (
            df__.groupby(["chi"])[[x for x in df__.columns if "Condition" in x]]
            .count()
            .mean(axis=1)
        )
        no = (
            df__.groupby(["chi"])[[x for x in df__.columns if "Operation" in x]]
            .count()
            .sum(axis=1)
        )
        df__.drop(
            columns=[
                x
                for x in df__.columns
                if "Date" in x or "Operation" in x or "Condition" in x
            ],
            inplace=True,
        )
        df__ = pd.DataFrame()
        df__["days_in_hospital"] = dih
        df__["average_number_conditions"] = nc
        df__["total_number_operations"] = no
        df__["number_admissions"] = number_stays

        # merge the third file
        df = df.merge(df__, how="inner", on="chi", suffixes=(False, False))

        # Final processing after merging
        df["death"] = [1 if x else 0 for x in pd.notna(df.date_of_death)]
        df["age"] = df.apply(find_age, axis=1).astype("int64")
        df.drop(columns=["date_of_birth", "date_of_death"], inplace=True)

        # save the dataframe
        df.to_csv(os.path.join(data_folder, "RDMP", processed_data_file))

    labels = df["death"]
    df.drop(columns=["death"], inplace=True)

    # OneHotEncoder
    for col in df.columns:
        if df[col].dtypes in ("bool", "object"):
            encoder = LabelEncoder()
            df[col] = encoder.fit_transform(df[col].values)

    return (df, labels)



"""Prompt to generate valid attack config."""

from __future__ import annotations

import logging

import yaml

from sacroml.attacks import factory
from sacroml.config import utils


def _get_defaults(name: str) -> dict:
    """Return an attack parameters and their defaults."""
    attack = factory.create_attack(name)
    return attack.get_params()


def _prompt_for_params(params: dict) -> None:
    """Prompt user to change parameter values."""
    for key, val in params.items():
        print(f"The current value for '{key}' is {val}.")
        if utils.get_bool("Do you want to change it?"):
            while True:
                new_val = input(f"Enter new value for '{key}': ").strip()
                try:
                    params[key] = type(val)(new_val)
                    break
                except ValueError:
                    print(f"Please enter a value of type {type(val).__name__}.")


def _get_attack(name: str) -> dict:
    """Get an attack configuration."""
    params: dict = _get_defaults(name)
    if not utils.get_bool("Use all defaults?"):
        _prompt_for_params(params)
    return {"name": name, "params": params}


def _prompt_for_attacks() -> list[dict]:
    """Prompt user for individual attack configurations."""
    attacks: list[dict] = []
    names: list[str] = list(factory.registry.keys())
    while utils.get_bool("Would you like to add an attack?"):
        while True:
            print(f"Attacks available: {', '.join(names)}")
            name: str = input("Which attack?: ")
            if name in names:
                attack = _get_attack(name)
                attacks.append(attack)
                print(f"{name} attack added.")
                break
            print("Please enter one of the available attacks.")
    return attacks


def _default_config() -> list[dict]:
    """Return a default configuration with all attacks."""
    attacks: list[dict] = []
    names: list[str] = list(factory.registry.keys())
    for name in names:
        params: dict = _get_defaults(name)
        attack: dict = {"name": name, "params": params}
        attacks.append(attack)
    return attacks


def prompt_for_attack() -> None:
    """Prompt user for information to generate attack config."""
    logging.disable(logging.ERROR)  # suppress info/warnings

    # get attack configuration
    if utils.get_bool("Generate default config with all attacks?"):
        attacks = _default_config()
    else:
        attacks = _prompt_for_attacks()

    # write to file
    filename: str = "attack.yaml"
    with open(filename, "w", encoding="utf-8") as fp:
        yaml.dump({"attacks": attacks}, fp)
    print(f"{filename} has been generated.")



"""Main entry point to sacroml."""

from __future__ import annotations

import argparse
import os

from sacroml.attacks.factory import run_attacks
from sacroml.config.attack import prompt_for_attack
from sacroml.config.target import prompt_for_target


def main() -> None:
    """Load target and attack configurations and run attacks."""
    parser = argparse.ArgumentParser(
        description="CLI for running attacks and generating config files"
    )

    subparsers = parser.add_subparsers(dest="cmd", required=True)

    run = subparsers.add_parser("run", help="Run attacks from YAML config files")

    run.add_argument("target_dir", type=str, help="Directory containing target.yaml")
    run.add_argument("attack_yaml", type=str, help="Attack YAML config")

    subparsers.add_parser("gen-target", help="Generate Target YAML config")
    subparsers.add_parser("gen-attack", help="Generate Attack YAML config")

    args = parser.parse_args()

    if args.cmd == "run":
        if os.path.isdir(args.target_dir) and os.path.isfile(args.attack_yaml):
            run_attacks(args.target_dir, args.attack_yaml)
        else:
            print("Both files must exist to run attacks.")
    elif args.cmd == "gen-target":
        prompt_for_target()
    elif args.cmd == "gen-attack":
        prompt_for_attack()


if __name__ == "__main__":  # pragma:no cover
    main()



"""Likelihood testing scenario from https://arxiv.org/pdf/2112.03570.pdf."""

from __future__ import annotations

import contextlib
import logging

import numpy as np
import sklearn
from fpdf import FPDF
from scipy.stats import norm, shapiro

from sacroml import metrics
from sacroml.attacks import report
from sacroml.attacks.attack import Attack
from sacroml.attacks.target import Target

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

EPS = 1e-16  # Used to avoid numerical issues


class LIRAAttack(Attack):
    """The main LiRA Attack class."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        output_dir: str = "outputs",
        write_report: bool = True,
        n_shadow_models: int = 100,
        p_thresh: float = 0.05,
        mode: str = "offline",
        fix_variance: bool = False,
        report_individual: bool = False,
    ) -> None:
        """Construct an object to execute a LiRA attack.

        Parameters
        ----------
        output_dir : str
            Name of the directory where outputs are stored.
        write_report : bool
            Whether to generate a JSON and PDF report.
        n_shadow_models : int
            Number of shadow models to be trained.
        p_thresh : float
            Threshold to determine significance of things. For instance
            auc_p_value and pdif_vals.
        mode : str
            Attack mode: {"offline", "offline-carlini", "online-carlini"}
        fix_variance : bool
            Whether to use the global standard deviation or per record.
        report_individual : bool
            Whether to report metrics for each individual record.
        """
        super().__init__(output_dir=output_dir, write_report=write_report)
        self.n_shadow_models: int = n_shadow_models
        self.p_thresh: float = p_thresh
        self.mode: str = mode
        self.fix_variance: bool = fix_variance
        self.report_individual: bool = report_individual

        self.result: dict = {}  # individual record results
        if self.report_individual:
            self.result["score"] = []
            self.result["label"] = []
            self.result["target_logit"] = []
            self.result["out_p_norm"] = []
            self.result["out_prob"] = []
            self.result["out_mean"] = []
            self.result["out_std"] = []
            if self.mode == "online-carlini":
                self.result["in_prob"] = []
                self.result["in_mean"] = []
                self.result["in_std"] = []

    def __str__(self):
        """Return the name of the attack."""
        return "LiRA Attack"

    def attack(self, target: Target) -> dict:
        """Run a LiRA attack from a Target object and a target model.

        Parameters
        ----------
        target : attacks.target.Target
            target as an instance of the Target class.

        Returns
        -------
        dict
            Attack report.
        """
        # prepare
        shadow_clf = sklearn.base.clone(target.model)
        target = self._check_and_update_dataset(target)
        # execute attack
        self._run(
            shadow_clf,
            target.X_train,
            target.y_train,
            target.model.predict_proba(target.X_train),
            target.X_test,
            target.y_test,
            target.model.predict_proba(target.X_test),
        )
        # create the report
        output = self._make_report(target)
        # write the report
        self._write_report(output)
        # return the report
        return output

    def _check_and_update_dataset(self, target: Target) -> Target:
        """Check that it is safe to use class variables to index prediction arrays.

        This has two steps:
        1. Replacing the values in y_train with their position in
        target.model.classes (will normally result in no change)
        2. Removing from the test set any rows corresponding to classes that
        are not in the training set.
        """
        y_train_new = []
        classes = list(target.model.classes_)
        for y in target.y_train:
            y_train_new.append(classes.index(y))
        target.y_train = np.array(y_train_new, int)
        logger.info(
            "new y_train has values and counts: %s",
            np.unique(target.y_train, return_counts=True),
        )
        ok_pos = []
        y_test_new = []
        for i, y in enumerate(target.y_test):
            if y in classes:
                ok_pos.append(i)
                y_test_new.append(classes.index(y))
        if len(y_test_new) != len(target.X_test):
            target.X_test = target.X_test[ok_pos, :]
        target.y_test = np.array(y_test_new, int)
        logger.info(
            "new y_test has values and counts: %s",
            np.unique(target.y_test, return_counts=True),
        )
        return target

    def _run(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        shadow_clf: sklearn.base.BaseEstimator,
        X_train: np.ndarray,
        y_train: np.ndarray,
        proba_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
        proba_test: np.ndarray,
    ) -> None:
        """Run the likelihood test.

        See p.6 (top of second column) for details.

        With mode "offline", we measure the probability of observing a
        confidence as high as the target model's under the null-hypothesis that
        the target point is a non-member. That is we, use the norm CDF.

        With mode "offline-carlini", we measure the probability that a target point
        did not come from the non-member distribution. That is, we use Carlini's
        implementation with a single norm (log) PDF.

        With mode "online-carlini", we use Carlini's implementation of the standard
        likelihood ratio test, measuring the ratio of probabilities the sample came
        from the two distributions. That is, the (log) PDF of pr_in minus pr_out.

        Parameters
        ----------
        shadow_clf : sklearn.Model
            An sklearn classifier that will be trained to form the shadow models.
            All hyperparameters should have been set.
        X_train : np.ndarray
            Data that was used to train the target model.
        y_train : np.ndarray
            Labels that were used to train the target model.
        proba_train : np.ndarray
            Array of predictions produced by the target model on the training data.
        X_test : np.ndarray
            Data that will be used to train the shadow models.
        y_test : np.ndarray
            Labels that will be used to train the shadow models.
        proba_test : np.ndarray
            Array of predictions produced by the target model on the shadow data.
        """
        logger.info("Running %s LiRA, fix_variance=%s", self.mode, self.fix_variance)

        # Combine target and shadow train, from which to sample datasets
        n_train_rows, _ = X_train.shape
        n_shadow_rows, _ = X_test.shape
        combined_x_train = np.vstack((X_train, X_test))
        combined_y_train = np.hstack((y_train, y_test))
        combined_target_preds = np.vstack((proba_train, proba_test))

        # Get the confidences of samples when in and not in the training set
        out_conf, in_conf = self._train_shadow_models(
            shadow_clf,
            combined_x_train,
            combined_y_train,
            n_train_rows,
        )

        # Get the LiRA scores, and how many confidences were normally distributed
        mia_scores, n_normal = self._compute_scores(
            combined_y_train, combined_target_preds, out_conf, in_conf
        )

        # Save metrics
        mia_clf = self._DummyClassifier()
        mia_scores = np.array(mia_scores)
        mia_labels = np.array([1] * n_train_rows + [0] * n_shadow_rows)
        y_pred_proba = mia_clf.predict_proba(mia_scores)
        self.attack_metrics = [metrics.get_metrics(y_pred_proba, mia_labels)]
        self.attack_metrics[-1]["n_normal"] = n_normal / (n_train_rows + n_shadow_rows)
        if self.report_individual:
            self.result["score"] = [score[1] for score in mia_scores]
            self.result["member"] = mia_labels
            self.attack_metrics[-1]["individual"] = self.result

        logger.info("Finished scenario")

    def _compute_scores(  # pylint: disable=too-many-locals
        self,
        combined_y_train: np.ndarray,
        combined_target_preds: np.ndarray,
        out_conf: dict[list[float]],
        in_conf: dict[list[float]],
    ) -> tuple[list[list[float]], int]:
        """Compute LiRA scores for each record."""
        logger.info("Computing scores")

        mia_scores: list[list[float]] = []
        n_normal: int = 0
        global_in_std: float = self._get_global_std(in_conf)
        global_out_std: float = self._get_global_std(out_conf)

        # score each record in the member and non-member sets
        for i, label in enumerate(combined_y_train):
            # get the target model behaviour on the record
            target_logit: float = _logit(combined_target_preds[i, label])
            # get behaviour observed with the record as a non-member
            out_mean, out_std = self._describe_conf(out_conf[i], global_out_std)
            # get behaviour observed with the record as a member
            in_mean, in_std = self._describe_conf(in_conf[i], global_in_std)
            # compare behaviour
            if self.mode == "offline":
                pr_out = norm.cdf(target_logit, loc=out_mean, scale=out_std + EPS)
                pr_in = 1 - pr_out
            elif self.mode == "online-carlini":
                pr_out = -norm.logpdf(target_logit, out_mean, out_std + EPS)
                pr_in = -norm.logpdf(target_logit, in_mean, in_std + EPS)
                # ratio
                pr_in = pr_in - pr_out
                pr_out = -pr_in
            elif self.mode == "offline-carlini":
                pr_out = -norm.logpdf(target_logit, out_mean, out_std + EPS)
                pr_in = -pr_out
            else:
                raise ValueError(f"Unsupported LiRA mode: {self.mode}")
            mia_scores.append([pr_in, pr_out])
            # test the non-member samples for normality
            out_p_norm = self._get_p_normal(np.array(out_conf[i]))
            if out_p_norm <= 0.05:
                n_normal += 1
            # save individual record result
            if self.report_individual:
                self.result["label"].append(label)
                self.result["target_logit"].append(target_logit)
                self.result["out_p_norm"].append(out_p_norm)
                self.result["out_prob"].append(pr_out)
                self.result["out_mean"].append(out_mean)
                self.result["out_std"].append(out_std + EPS)
                if self.mode == "online-carlini":
                    self.result["in_prob"].append(pr_in)
                    self.result["in_mean"].append(in_mean)
                    self.result["in_std"].append(in_std + EPS)
        return mia_scores, n_normal

    def _describe_conf(
        self, confidences: list[float], global_std: float
    ) -> tuple[float, float]:
        """Return the mean and standard deviation of a list of confidences."""
        scores: np.ndarray = np.array(confidences)
        mean: float = 0
        std: float = 0
        if not np.isnan(scores).all():
            mean = np.nanmean(scores)
            std = np.nanstd(scores)
        if self.fix_variance:
            std = global_std
        return mean, std

    def _get_global_std(self, confidences: dict[str, list[float]]) -> float:
        """Return the global standard deviation."""
        global_std: float = 0
        if self.fix_variance:
            # requires conversion from a dict of diff size proba lists
            arrays = list(confidences.values())
            combined = np.concatenate(arrays)
            if not np.isnan(combined).all():
                global_std = np.nanstd(combined)
        return global_std

    def _get_p_normal(self, samples: np.ndarray) -> float:
        """Test whether a set of samples is normally distributed."""
        p_normal: float = np.NaN
        if np.nanvar(samples) > EPS:
            with contextlib.suppress(ValueError):
                _, p_normal = shapiro(samples)
        return p_normal

    def _train_shadow_models(  # pylint: disable=too-many-locals
        self,
        shadow_clf: sklearn.base.BaseEstimator,
        combined_x_train: np.ndarray,
        combined_y_train: np.ndarray,
        n_train_rows: int,
    ) -> tuple[dict, dict]:
        """Train shadow models and return confidence scores.

        Parameters
        ----------
        shadow_clf : sklearn.base.BaseEstimator
            An sklearn classifier that will be trained to form the shadow models.
        combined_x_train : np.ndarray
            Array of combined train and test features.
        combined_y_train : np.ndarray
            Array of combined train and test labels.
        n_train_rows : int
            Number of samples in the training set.

        Returns
        -------
        tuple[dict, dict]
            Dictionary of confidences when not in the training set.
            Dictionary of confidences when in the training set.
        """
        logger.info("Training shadow models")

        n_combined, _ = combined_x_train.shape
        out_conf: dict = {i: [] for i in range(n_combined)}
        in_conf: dict = {i: [] for i in range(n_combined)}
        indices: np.ndarray = np.arange(0, n_combined, 1)

        for model_idx in range(self.n_shadow_models):
            if model_idx % 10 == 0:
                logger.info("Trained %d models", model_idx)
            # Pick the indices to use for training this one
            np.random.seed(model_idx)  # Reproducibility
            these_idx = np.random.choice(indices, n_train_rows, replace=False)
            # Fit the shadow model
            shadow_clf.set_params(random_state=model_idx)
            shadow_clf.fit(
                combined_x_train[these_idx, :],
                combined_y_train[these_idx],
            )
            # map a class to a column
            class_map = {c: i for i, c in enumerate(shadow_clf.classes_)}
            # generate shadow confidences
            shadow_confidences = shadow_clf.predict_proba(combined_x_train)
            these_idx = set(these_idx)
            for i, conf in enumerate(shadow_confidences):
                # logit of the correct class
                label = class_map.get(combined_y_train[i], -1)
                # Occasionally, the random data split will result in classes being
                # absent from the training set. In these cases label will be -1 and
                # we include logit(0) instead of discarding
                logit = _logit(0) if label < 0 else _logit(conf[label])
                if i not in these_idx:
                    out_conf[i].append(logit)
                else:
                    in_conf[i].append(logit)
        return out_conf, in_conf

    def _construct_metadata(self) -> None:
        """Construct the metadata object."""
        super()._construct_metadata()

        pdif = np.exp(-self.attack_metrics[0]["PDIF01"])

        self.metadata["global_metrics"]["PDIF_sig"] = (
            f"Significant at p={self.p_thresh}"
            if pdif <= self.p_thresh
            else f"Not significant at p={self.p_thresh}"
        )

        auc_p, auc_std = metrics.auc_p_val(
            self.attack_metrics[0]["AUC"],
            self.attack_metrics[0]["n_pos_test_examples"],
            self.attack_metrics[0]["n_neg_test_examples"],
        )
        self.metadata["global_metrics"]["AUC_sig"] = (
            f"Significant at p={self.p_thresh}"
            if auc_p <= self.p_thresh
            else f"Not significant at p={self.p_thresh}"
        )
        self.metadata["global_metrics"]["null_auc_3sd_range"] = (
            f"{0.5 - 3 * auc_std} -> {0.5 + 3 * auc_std}"
        )

    def _make_pdf(self, output: dict) -> FPDF:
        """Create PDF report."""
        return report.create_lr_report(output)

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the metadata object after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}
        for rep, _ in enumerate(self.attack_metrics):
            attack_metrics_instances["instance_" + str(rep)] = self.attack_metrics[rep]
        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        return attack_metrics_experiment

    class _DummyClassifier:
        """A Dummy Classifier to allow this code to work with get_metrics."""

        def predict(self, X_test):
            """Return an array of 1/0 depending on value in second column."""
            return 1 * (X_test[:, 1] > 0.5)

        def predict_proba(self, X_test):
            """Simply return the X_test."""
            return X_test


def _logit(p: float) -> float:
    """Return standard logit.

    Parameters
    ----------
    p : float
        value to evaluate logit at.

    Returns
    -------
    float
        logit(p)

    Notes
    -----
    If `p` is close to 0 or 1, evaluating the log will result in numerical
    instabilities. This code thresholds `p` at `EPS` and `1 - EPS` where `EPS`
    defaults at 1e-16.
    """
    if p > 1 - EPS:
        p = 1 - EPS
    p = max(p, EPS)
    return np.log(p / (1 - p))



"""Template for implementing supplementary models.

Obviously we have invented an sklearn ensemble called ModelToMakeSafer. Replace
this with details of the model you wish to create a wrapper for and then remove
the comment which disables the pylint warning.
"""

# pylint: disable=duplicate-code

from __future__ import annotations

import copy

import numpy as np
from dictdiffer import diff
from sklearn.ensemble import ModelToMakeSafer  # pylint: disable=E0611
from sklearn.tree import DecisionTreeClassifier

from sacroml.safemodel.safemodel import SafeModel

from .safedecisiontreeclassifier import decision_trees_are_equal


def check_present(
    item: str, curr_separate: dict, saved_separate: dict
) -> tuple[str, bool]:
    """Check item is present in both dicts and report suitably."""
    disclosive = False
    msg = ""
    if curr_separate[item] == "Absent" and saved_separate[item] == "Absent":
        disclosive = True
        msg += "Error: model has not been fitted to data.\n"

    elif curr_separate[item] == "Absent":
        disclosive = True
        msg += "Error: current version of model has had trees removed after fitting.\n"

    elif saved_separate[item] == "Absent":
        disclosive = True
        msg += "Error: current version of model has had trees manually edited.\n"
    return msg, disclosive


class SafeModelToMakeSafe(SafeModel, ModelToMakeSafer):
    """Privacy protected ModelToMakeSafer."""

    def __init__(self, **kwargs: dict) -> None:
        """Create model and apply constraints to params."""
        SafeModel.__init__(self)
        self.k_anonymity = 0
        self.basemodel_paramnames = [
            "edit",
            "this",
            "list",
            "to",
            "contain",
            "just",
            "the",
            "valid",
            "parameters",
            "for",
            "the",
            "class",
            "you ",
            "are",
            "creating",
            "a",
            "safe",
            "wrapper",
            "version",
            "of",
        ]

        the_kwds = {}
        for key, val in kwargs.items():
            if key in self.basemodel_paramnames:
                the_kwds[key] = val
        ModelToMakeSafer.__init__(self, **the_kwds)
        self.model_type: str = "ModelToMakeSafer"
        super().preliminary_check(apply_constraints=True, verbose=True)
        self.ignore_items = [
            "model_save_file",
            "ignore_items",
            "base_estimator_",
            "timestamp",
        ]
        self.examine_seperately_items = ["base_estimator", "estimators_"]

    def additional_checks(  # pylint: disable=too-many-nested-blocks,too-many-branches
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, str]:
        """Perform model specific checks.

        This example shows how to deal with instances of sklearn's tree class
        as base estimators in a forest (line 99)
        or as single estimators (lines 114-118).

        Notes
        -----
        Call the super function to deal with any items that are lists. For example:
        >>> msg, disclosive = super().additional_checks(curr_separate, saved_separate)
        """
        msg = ""
        disclosive = False
        # now the relevant ModelToMakeSafer specific things
        for item in self.examine_seperately_items:
            if item == "base_estimator":
                try:
                    the_type = type(self.base_estimator)
                    if not isinstance(self.saved_model["base_estimator_"], the_type):
                        msg += "Warning: model was fitted with different base estimator type.\n"
                        disclosive = True
                except AttributeError:
                    msg += "Error: model has not been fitted to data.\n"
                    disclosive = True

            elif item == "estimators_":
                msg2, disclosive2 = check_present(item, curr_separate, saved_separate)
                msg += msg2
                if disclosive2:
                    disclosive = True
                else:
                    try:
                        num1 = len(curr_separate[item])
                        num2 = len(saved_separate[item])
                        if num1 != num2:
                            msg += (
                                f"Fitted model has {num2} estimators "
                                f"but requested version has {num1}.\n"
                            )
                            disclosive = False
                        else:
                            for idx in range(num1):
                                (
                                    same,
                                    msg2,
                                ) = decision_trees_are_equal(
                                    curr_separate[item][idx], saved_separate[item][idx]
                                )
                                if not same:
                                    disclosive = True
                                    msg += f"Forest base estimators {idx} differ."
                                    msg += msg2

                    except BaseException as error:  # pylint:disable=broad-except
                        msg += (
                            "In Safe ModelToMakeSafer.additional_checks: "
                            f"Unable to check {item} as an exception occurred: {error}.\n"
                        )
                        same = False

            elif isinstance(curr_separate[item], DecisionTreeClassifier):
                diffs_list = list(diff(curr_separate[item], saved_separate[item]))
                if len(diffs_list) > 0:
                    disclosive = True
                    msg += f"structure {item} has {len(diffs_list)} differences: {diffs_list}"
        return msg, disclosive

    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model and store model dict."""
        super().fit(x, y)
        self.k_anonymity = self.get_k_anonymity(x)
        self.saved_model = copy.deepcopy(self.__dict__)

    def get_k_anonymity(self, x: np.ndarray) -> int:
        """Calculate the k-anonymity of a random forest model.

        The k-anonymity is the minimum of the anonymity for each record.
        That is defined as the size of the set of records which
        appear in the same leaf as the record in every tree.
        """
        # dataset must be 2-D
        assert len(x.shape) == 2

        num_records = x.shape[0]
        num_trees = self.n_estimators
        k_anon_val = np.zeros(num_records, dtype=int)

        # ending leaf node by record(row) and tree (column)
        all_leaves = np.zeros((num_records, num_trees), dtype=int)
        for this_tree in range(num_trees):
            this_leaves = self.estimators_[this_tree].apply(x)
            for record in range(num_records):
                all_leaves[record][this_tree] = this_leaves[record]

        for record in range(num_records):
            # start by assuming everything co-occurs
            appears_together = list(range(num_records))
            # iterate through trees
            for this_tree in range(num_trees):
                this_leaf = all_leaves[record][this_tree]

                together = copy.copy(appears_together)
                # removing records which go to other leaves
                for other_record in together:
                    if all_leaves[other_record][this_tree] != this_leaf:
                        appears_together.remove(other_record)

            k_anon_val[record] = len(appears_together)
        return k_anon_val.min()



"""Privacy protected Decision Tree classifier."""

from __future__ import annotations

import copy
from typing import Any

import numpy as np
from dictdiffer import diff
from sklearn.tree import DecisionTreeClassifier

from sacroml.safemodel.reporting import get_reporting_string
from sacroml.safemodel.safemodel import SafeModel


def decision_trees_are_equal(
    tree1: DecisionTreeClassifier, tree2: DecisionTreeClassifier
) -> tuple[bool, str]:
    """Compare two estimators of type sklearn.tree."""
    msg = ""
    same = True

    try:
        tree1_dict = copy.deepcopy(tree1.__dict__)
        tree1_tree = tree1_dict.pop("tree_", "Absent")
        tree2_dict = copy.deepcopy(tree2.__dict__)
        tree2_tree = tree2_dict.pop("tree_", "Absent")

        # comparison on list of "simple" parameters
        match = list(diff(tree1_dict, tree2_dict, expand=True))
        num_differences = len(match)
        if num_differences > 0:
            same = False
            msg += get_reporting_string(
                name="basic_params_differ", length=num_differences
            )
            for i in range(num_differences):
                if match[i][0] == "change":
                    msg += f"parameter {match[i][1]} changed from {match[i][2][1]} "
                    msg += f"to {match[i][2][0]}\n"
                else:
                    msg += f"{match[i]}\n"

        # now internal tree params
        same2, msg2 = decision_tree_internal_trees_are_equal(tree1_tree, tree2_tree)
        if same2 is False:
            same = False
            msg += msg2

    except BaseException as error:  # pylint:disable=broad-except  #pragma:no cover
        msg += get_reporting_string(name="unable_to_check", error=error)
        same = False

    return same, msg


def decision_tree_internal_trees_are_equal(
    tree1_tree: Any, tree2_tree: Any
) -> tuple[bool, str]:
    """Test for equality of the internal structures in a sklearn.tree._tree.

    For example, the structure, feature and threshold in each internal node etc.
    """
    same = True
    msg = ""
    tree_internal_att_names = (
        "capacity",
        "children_left",
        "children_right",
        "feature",
        "impurity",
        "max_depth",
        "n_node_samples",
        "node_count",
        "threshold",
        "value",
        "weighted_n_node_samples",
    )

    try:
        if tree1_tree == "Absent" and tree2_tree == "Absent":
            msg += get_reporting_string(name="neither_tree_trained")
            # "neither tree trained"
        elif tree1_tree == "Absent":
            msg += get_reporting_string(name="tree1_not_trained")
            # "tree1 not trained"
            same = False
        elif tree2_tree == "Absent":
            msg += get_reporting_string(name="tree2_not_trained")
            # "tree2 not trained"
            same = False
        else:
            for attr in tree_internal_att_names:
                t1val = getattr(tree1_tree, attr)
                t2val = getattr(tree2_tree, attr)
                if isinstance(t1val, np.ndarray):
                    if not np.array_equal(t1val, t2val):
                        msg += get_reporting_string(
                            name="internal_attribute_differs", attr=attr
                        )
                        same = False
                elif t1val != t2val:
                    msg += get_reporting_string(
                        name="internal_attribute_differs", attr=attr
                    )
                    same = False
    except BaseException as error:  # pylint:disable=broad-except #pragma:no cover
        msg += get_reporting_string(name="exception_occurred", error=error)
    return same, msg


def get_tree_k_anonymity(thetree: DecisionTreeClassifier, X: Any) -> int:
    """Return the smallest number of data items in any leaf."""
    leaves = thetree.apply(X)
    uniqs_counts = np.unique(leaves, return_counts=True)
    return np.min(uniqs_counts[1])


class SafeDecisionTreeClassifier(SafeModel, DecisionTreeClassifier):  # pylint: disable=too-many-ancestors
    """Privacy protected Decision Tree classifier."""

    def __init__(self, **kwargs: dict) -> None:
        """Create model and apply constraints to params."""
        SafeModel.__init__(self)
        self.basemodel_paramnames = [
            "criterion",
            "splitter",
            "max_depth",
            "min_samples_split",
            "min_samples_leaf",
            "min_weight_fraction_leaf",
            "max_features",
            "random_state",
            "max_leaf_nodes",
            "min_impurity_decrease",
            "class_weight",
            "ccp_alpha",
        ]

        the_kwds = {}
        for key, val in kwargs.items():
            if key in self.basemodel_paramnames:
                the_kwds[key] = val
        DecisionTreeClassifier.__init__(self, **the_kwds)
        self.model_type: str = "DecisionTreeClassifier"
        super().preliminary_check(apply_constraints=False, verbose=True)
        self.ignore_items = [
            "model_save_file",
            "basemodel_paramnames",
            "ignore_items",
            "timestamp",
        ]
        self.examine_seperately_items = ["tree_"]
        self.k_anonymity = 0

    def additional_checks(
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, str]:
        """Decision Tree-specific checks."""
        # call the super function to deal with any items that are lists
        # just in case we add any in the future
        msg, disclosive = super().additional_checks(curr_separate, saved_separate)
        # now deal with the decision-tree specific things
        # which for now means the attribute "tree_" which is a sklearn tree
        same, msg = decision_tree_internal_trees_are_equal(
            curr_separate["tree_"], saved_separate["tree_"]
        )
        if not same:
            disclosive = True
        if len(curr_separate) > 1:
            msg += get_reporting_string(name="unexpected_item")
        return msg, disclosive

    def fit(  # pylint: disable=arguments-differ
        self, x: np.ndarray, y: np.ndarray
    ) -> None:
        """Fit model and store k-anonymity and model dict."""
        super().fit(x, y)
        # calculate k-anonymity her since we have the tainigf data
        leaves = self.apply(x)
        uniqs_counts = np.unique(leaves, return_counts=True)
        self.k_anonymity = np.min(uniqs_counts[1])
        self.saved_model = copy.deepcopy(self.__dict__)



"""Utilities for config prompt generation."""

from __future__ import annotations

import os
import shutil

yes: list[str] = ["yes", "y", "yeah", "yep", "sure", "ok"]
no: list[str] = ["no", "n", "nope", "nah"]


def get_bool(prompt: str) -> bool:
    """Get a Boolean response to a prompt.

    Parameters
    ----------
    prompt : str
        Message to display to user.

    Returns
    -------
    bool
        Whether the user responded yes or no.
    """
    while True:
        response = input(prompt + " (yes/no): ").strip().lower()
        if response in yes:
            return True
        if response in no:
            return False
        print("Invalid input. Please enter 'yes' or 'no'.")


def check_dir(path: str) -> str:
    """Check directory exists and create as necessary.

    Parameters
    ----------
    path : str
        Directory to check exists.

    Returns
    -------
    str
        Directory, possibly changed by user if pre-existing.
    """
    if os.path.isdir(path):
        print(f"Directory '{path}' already exists.")
        resp = get_bool(
            "Continue using this directory and delete its current contents?"
        )
        if resp:
            shutil.rmtree(path)
        else:
            path = input("Specify an alternative directory: ")
            return check_dir(path)
    return path



"""Prompt to generate valid target config."""

from __future__ import annotations

import ast
import os
import sys

from sacroml.attacks.target import Target
from sacroml.config import utils

arrays_pro = ["X_train", "y_train", "X_test", "y_test"]
arrays_raw = ["X", "y", "X_train_orig", "y_train_orig", "X_test_orig", "y_test_orig"]
arrays_proba = ["proba_train", "proba_test"]
encodings = ["onehot", "str", "int", "float"]

MAX_FEATURES = 64  # maximum features to prompt


def _get_arrays(target: Target, arrays: list[str]) -> None:
    """Prompt user for the paths to array data."""
    for arr in arrays:
        while True:
            path = input(f"What is the path to {arr}? ")
            try:
                target.load_array(path, arr)
                break
            except FileNotFoundError:
                print("File does not exist. Please try again.")
            except ValueError as e:
                print(f"WARNING: {e}")
                break


def _get_dataset_name(target: Target) -> None:
    """Prompt user for the name of a dataset."""
    target.dataset_name = input("What is the name of the dataset? ")


def _get_feature_encoding(feat: int) -> str:
    """Prompt user for feature encoding."""
    while True:
        encoding = input(f"What is the encoding of feature {feat}? ")
        if encoding in encodings:
            return encoding
        print("Invalid encoding. Please try again.")


def _get_feature_indices(feat: int) -> list[int]:
    """Prompt user for feature indices."""
    while True:
        indices = input(f"What are the indices for feature {feat}, e.g., [1,2,3]? ")
        try:
            indices = ast.literal_eval(indices)
            if isinstance(indices, list) and all(isinstance(i, int) for i in indices):
                return indices
            print("Invalid input. Please enter a list of integers like [1,2,3].")
        except (ValueError, SyntaxError):
            print("Invalid input. Please enter a valid list of integers.")


def _get_features(target: Target) -> None:
    """Prompt user for dataset features."""
    print("To run attribute inference attacks the features must be described.")
    n_features = input("How many features does this dataset have? ")
    n_features = int(n_features)
    if n_features > MAX_FEATURES:
        print("There are too many features to add via prompt.")
        print("You can edit the 'target.yaml' to add features later.")
        print("Note: this is only necessary for attribute inference.")
        return
    print("The name, index, and encoding are needed for each feature.")
    print("For example: feature 0 = 'parents', '[0, 1, 2]', 'onehot'")
    if utils.get_bool("Do you want to add this information?"):
        print(f"Valid encodings: {', '.join(encodings)}")
        for i in range(n_features):
            name = input(f"What is the name of feature {i}? ")
            indices = _get_feature_indices(i)
            encoding = _get_feature_encoding(i)
            target.add_feature(name, indices, encoding)


def _get_model_path() -> str:
    """Prompt user for path to a saved fitted model."""
    if not utils.get_bool("Do you have a saved fitted model?"):
        print("Cannot generate a target config without a fitted model.")
        sys.exit()

    while True:
        path = input("Enter path including the full filename: ")
        if os.path.isfile(path):
            break
        print("File does not exist. Please try again.")
    return path


def _get_proba(target: Target) -> None:
    """Prompt user for model predicted probabilities."""
    if not utils.get_bool("Do you have the model predicted probabilities?"):
        print("Cannot run any attacks without a supported model or probabilities.")
        sys.exit()
    _get_arrays(target, arrays_proba)


def prompt_for_target() -> None:
    """Prompt user for information to generate target config."""
    target = Target()
    path: str = utils.check_dir("target")

    model_path: str = _get_model_path()
    try:  # attempt to load saved model
        target.load_model(model_path)
    except ValueError:  # unsupported model, require probas
        print("Unable to load model.")
        _get_proba(target)

    _get_dataset_name(target)
    if utils.get_bool("Do you know the paths to processed data?"):
        _get_arrays(target, arrays_pro)
        _get_features(target)
    if utils.get_bool("Do you know the paths to original raw data?"):
        _get_arrays(target, arrays_raw)
    target.save(path)
    print(f"Target generated in directory: '{path}'")



"""Privacy protected Random Forest classifier."""

from __future__ import annotations

import copy

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from sacroml.safemodel.reporting import get_reporting_string
from sacroml.safemodel.safemodel import SafeModel

from .safedecisiontreeclassifier import decision_trees_are_equal

# pylint: disable=too-many-ancestors, unidiomatic-typecheck


class SafeRandomForestClassifier(SafeModel, RandomForestClassifier):
    """Privacy protected Random Forest classifier."""

    def __init__(self, **kwargs: dict) -> None:
        """Create model and apply constraints to params."""
        SafeModel.__init__(self)
        self.basemodel_paramnames = [
            "n_estimators",
            "criterion",
            "max_depth",
            "min_samples_split",
            "min_samples_leaf",
            "min_weight_fraction_leaf",
            "max_features",
            "max_leaf_nodes",
            "min_impurity_decrease",
            "bootstrap",
            "oob_score",
            "n_jobs",
            "random_state",
            "verbose",
            "warm_start",
            "class_weight",
            "ccp_alpha",
            "max_samples",
        ]

        the_kwds = {}
        for key, val in kwargs.items():
            if key in self.basemodel_paramnames:
                the_kwds[key] = val
        RandomForestClassifier.__init__(self, **the_kwds)
        self.model_type: str = "RandomForestClassifier"
        super().preliminary_check(apply_constraints=False, verbose=True)
        self.ignore_items = [
            "model_save_file",
            "ignore_items",
            "estimator_",  # this is an object
            "timestamp",
        ]
        self.examine_seperately_items = ["estimators_", "estimator"]
        self.k_anonymity = 0

    def additional_checks(  # pylint: disable=too-many-nested-blocks
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, str]:
        """Perform Random Forest specific checks.

        NOTE: this is never called if the model has not been fitted.
        """
        msg = ""
        disclosive = False
        # now the relevant random-forest specific things
        for item in self.examine_seperately_items:
            # template for class of things that make up forest
            if item == "estimator":
                if type(curr_separate[item]) != type(saved_separate[item]):
                    msg += get_reporting_string(
                        name="param_changed_from_to",
                        key="estimator",
                        val=saved_separate[item],
                        cur_val=curr_separate[item],
                    )
                    disclosive = True
            # the forest itself
            elif item == "estimators_":
                try:
                    num1 = len(curr_separate[item])
                    num2 = len(saved_separate[item])
                    if num1 != num2:
                        msg += get_reporting_string(
                            name="different_num_estimators", num1=num1, num2=num2
                        )
                        disclosive = True
                    else:
                        changed = False
                        num_diff_trees = 0
                        for idx in range(num1):
                            same, _ = decision_trees_are_equal(
                                curr_separate[item][idx], saved_separate[item][idx]
                            )
                            if not same:
                                changed = True
                                num_diff_trees += 1
                        if changed:
                            msg += get_reporting_string(
                                name="forest_estimators_differ", idx=num_diff_trees
                            )
                            disclosive = True
                except (
                    BaseException  # pylint: disable=broad-except
                ) as error:  # pragma:no coverion
                    msg += get_reporting_string(
                        name="unable_to_check_item", item=item, error=error
                    )
                    same = False

        return msg, disclosive

    # pylint: disable=arguments-differ
    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model and store model dict."""
        super().fit(x, y)
        self.k_anonymity = self.get_k_anonymity(x)
        self.saved_model = copy.deepcopy(self.__dict__)

    def get_k_anonymity(self, x: np.ndarray) -> int:
        """Calculate the k-anonymity of a random forest model.

        The k-anonymity is the minimum of the anonymity for each record.
        That is defined as the size of the set of records which
        appear in the same leaf as the record in every tree.
        """
        # dataset must be 2-D
        assert len(x.shape) == 2

        num_records = x.shape[0]
        num_trees = self.n_estimators
        k_anon_val = np.zeros(num_records, dtype=int)

        # ending leaf node by record(row) and tree (column)
        all_leaves = np.zeros((num_records, num_trees), dtype=int)
        for this_tree in range(num_trees):
            this_leaves = self.estimators_[this_tree].apply(x)
            for record in range(num_records):
                all_leaves[record][this_tree] = this_leaves[record]

        for record in range(num_records):
            # start by assuming everything co-occurs
            appears_together = list(range(num_records))
            # iterate through trees
            for this_tree in range(num_trees):
                this_leaf = all_leaves[record][this_tree]

                together = copy.copy(appears_together)
                # removing records which go to other leaves
                for other_record in together:
                    if all_leaves[other_record][this_tree] != this_leaf:
                        appears_together.remove(other_record)

            k_anon_val[record] = len(appears_together)
        return k_anon_val.min()



"""Methods to producte standard reporting strings."""


def get_reporting_string(**kwargs: dict) -> str:
    """Return a standard formatted string from a dictionary of f-strings.

    Parameters
    ----------
    name : string
         The dictionary key and the name of the string to return.
    all-the-keywords : Any Type
         one or more values expected in the f-string.

    Returns
    -------
    msg : string
         A standard message string.

    Notes
    -----
    Sometimes an f-string has no parameters.
    Sometimes there are multiple parameters embedded in the f-string.
    """
    the_kwargs = kwargs

    # check for the name used to reference the diction ary of f-strings
    # if no name is found return an error string

    inter_params = {
        "attr": None,
        "batch_size": None,
        "current_epsilon": None,
        "cur_val": None,
        "diffs_list": "",
        "e": None,
        "epochs": None,
        "er": None,
        "error": None,
        "idx": None,
        "item": None,
        "key": None,
        "layer": None,
        "length": None,
        "match": "",
        "name": None,
        "num1": None,
        "num2": None,
        "num_samples": None,
        "model_type": None,
        "msg": None,
        "operator": None,
        "optimizer": None,
        "opt_msg": None,
        "suffix": None,
        "v1": None,
        "v2": None,
        "val": None,
    }
    # initialise all the values from the keyword parameters
    # check for parameters in keyword arguments.
    # if keyword is found set dictionary item value to that
    # provided by the keyword
    # if keyword is not found dictionary item value remains None,
    # and an error message is returned.

    for param, value in the_kwargs.items():
        inter_params[param] = value

    if inter_params["name"] is None:
        return "Error - get_reporting_string: No 'name' given"

    # A dictionary of f-strings follows

    report_string = {
        "NULL": (""),
        "less_than_min_value": (
            f"- parameter {inter_params['key']} = {inter_params['cur_val']}"
            f" identified as less than the recommended min value of {inter_params['val']}."
        ),
        "greater_than_max_value": (
            f"- parameter {inter_params['key']} = {inter_params['cur_val']}"
            f" identified as greater than the recommended max value of {inter_params['val']}."
        ),
        "different_than_fixed_value": (
            f"- parameter {inter_params['key']} = {inter_params['cur_val']}"
            f" identified as different than the recommended fixed value of {inter_params['val']}."
        ),
        "different_than_recommended_type": (
            f"- parameter {inter_params['key']} = {inter_params['cur_val']}"
            f" identified as different type to recommendation of {inter_params['val']}.\n"
        ),
        "change_param_type": (
            f"\nChanged parameter type for {inter_params['key']} to {inter_params['val']}.\n"
        ),
        "not_implemented_for_change": (
            f"Nothing currently implemented to change type of parameter {inter_params['key']} "
            f"from {type(inter_params['cur_val']).__name__} to {inter_params['val']}.\n"
        ),
        "changed_param_equal": (
            f"\nChanged parameter {inter_params['key']} = {inter_params['val']}.\n"
        ),
        "unknown_operator": (
            f"- unknown operator in parameter specification {inter_params['operator']}"
        ),
        "warn_possible_disclosure_risk": (
            "WARNING: model parameters may present a disclosure risk:\n"
        ),
        "within_recommended_ranges": (
            "Model parameters are within recommended ranges.\n"
        ),
        "error_not_called_fit": (
            "Error: user has not called fit() method or has deleted saved values."
        ),
        "recommend_do_not_release": ("Recommendation: Do not release."),
        "layer_configs_differ": (
            f"Layer {inter_params['layer']} configs "
            f"differ in {inter_params['length']} places:\n"
        ),
        "error_reloading_model_v1": (
            f"Error re-loading  model from {inter_params['v1']}:  {inter_params['e']}"
        ),
        "error_reloading_model_v2": (
            f"Error re-loading  model from {inter_params['v2']}: {inter_params['e']}"
        ),
        "same_ann_config": ("configurations match"),
        "different_layer_count": ("models have different numbers of layers"),
        "batch_size_zero": (
            "Batch size of 0 not allowed"
            "setting batch_size =32.\n"
            "Alter self.batch_size manually if required."
        ),
        "division_by_zero": ("Division by zero setting batch_size =1"),
        "dp_requirements_met": (
            "The requirements for DP are met, "
            f"current epsilon is: {inter_params['current_epsilon']}."
            "Calculated from the parameters:  "
            f"Num Samples = {inter_params['num_samples']}, "
            f"batch_size = {inter_params['batch_size']}, epochs = {inter_params['epochs']}.\n"
        ),
        "dp_requirements_not_met": (
            f"The requirements for DP are not met, "
            f"current epsilon is: {inter_params['current_epsilon']}.\n"
            f"To attain recommended DP the following parameters can be changed:  "
            f"Num Samples = {inter_params['num_samples']},"
            f"batch_size = {inter_params['batch_size']},"
            f"epochs = {inter_params['epochs']}.\n"
        ),
        "basic_params_differ": (
            "Warning: basic parameters differ in " f"{inter_params['length']} places:\n"
        ),
        "param_changed_from_to": (
            f"parameter {inter_params['key']} changed from {inter_params['val']} "
            f"to {inter_params['cur_val']} after model was fitted.\n"
        ),
        "unable_to_check": (
            f"Unable to check as an exception occurred: {inter_params['error']}"
        ),
        "neither_tree_trained": ("neither tree trained"),
        "tree1_not_trained": ("tree1 not trained"),
        "tree2_not_trained": ("tree2 not trained"),
        "internal_attribute_differs": (
            f"internal tree attribute {inter_params['attr']} differs\n"
        ),
        "exception_occurred": (f"An exception occurred: {inter_params['error']}"),
        "unexpected_item": (
            "unexpected item in curr_seperate dict "
            " passed by generic additional checks."
        ),
        "warn_fitted_different_base": (
            "Warning: model was fitted with different base estimator type.\n"
        ),
        "error_model_not_fitted": ("Error: model has not been fitted to data.\n"),
        "current_item_removed": (
            f"Error, item  {inter_params['item']} "
            "present in  saved but not current model.\n"
        ),
        "saved_item_removed": (
            f"Error, item  {inter_params['item']} "
            "present in  current but not saved model.\n"
        ),
        "both_item_removed": f"Note that item {inter_params['item']} missing from both versions.\n",
        "trees_edited": (
            "Error: current version of model has had trees manually edited.\n"
        ),
        "different_num_estimators": (
            f"Fitted model has {inter_params['num2']} estimators "
            f"but requested version has {inter_params['num1']}.\n"
        ),
        "forest_estimators_differ": (
            f"{inter_params['idx']} forest base estimators have been changed.\n"
        ),
        "unable_to_check_item": (
            "In SafeRandomForest.additional_checks: "
            f"Unable to check {inter_params['item']} as an exception occurred:"
            f"{inter_params['error']}.\n"
        ),
        "structure_differences": (
            f"structure {inter_params['item']} has "
            f"{len(inter_params['diffs_list'])} differences: "
            f"{inter_params['diffs_list']}"
        ),
        "no_dp_gradients_key": (
            "optimizer does not contain key _was_dp_gradients_called" " so is not DP."
        ),
        "found_dp_gradients_key": (
            "optimizer does  contain key _was_dp_gradients_called" " so should be DP."
        ),
        "changed_opt_no_fit": (
            "optimizer has been changed but fit() has not been rerun."
        ),
        "dp_optimizer_run": (
            " value of optimizer._was_dp_gradients_called is True, "
            "so DP variant of optimizer has been run"
        ),
        "unrecognised_combin ation": ("unrecognised combination"),
        "optimizer_allowed": (f"optimizer {inter_params['optimizer']} allowed"),
        "optimizer_not_allowed": (
            f"optimizer {inter_params['optimizer']} is not allowed"
        ),
        "using_dp_sgd": ("Changed parameter optimizer = 'DPKerasSGDOptimizer'"),
        "using_dp_adagrad": ("Changed parameter optimizer = 'DPKerasAdagradOptimizer'"),
        "using_dp_adam": ("Changed parameter optimizer = 'DPKerasAdamOptimizer'"),
        "during_compilation": (f"During compilation: {inter_params['opt_msg']}"),
        "recommend_not_release": (
            f"Recommendation is not to release because {inter_params['msg']}.\n"
        ),
        "error_saving_file": (
            f"saving as a {inter_params['suffix']} "
            f"file gave this error message:  {inter_params['er']}"
        ),
        "loading_from_unsupported": (
            f"loading from a {inter_params['suffix']} "
            "file is currently not supported"
        ),
        "opt_config_changed": ("Optimizer config has been changed since training."),
        "epsilon_above_normal": (
            f"WARNING: epsilon {inter_params['current_epsilon']} "
            "is above normal max recommended value.\n"
            "Discussion with researcher needed.\n"
        ),
        "recommend_further_discussion": (
            f"Recommendation is further discussion needed " f"{inter_params['msg']}.\n"
        ),
        "recommend_allow_release": ("Recommendation is to allow release.\n"),
        "allow_release_eps_below_max": (
            "Recommendation: Allow release.\n"
            f"Epsilon vale of model {inter_params['current_epsilon']} "
            "is below normal max recommended value.\n"
        ),
        "input_filename_with_extension": (
            "Please input a name with extension for the model to be saved."
        ),
        "filename_must_indicate_type": ("file name must indicate type as a suffix"),
        "suffix_not_supported_for_type": (
            f"{inter_params['suffix']} file suffix  not supported "
            f"for models of type {inter_params['model_type']}.\n"
        ),
    }

    # return the correct formatted string

    return report_string[inter_params["name"]]



"""Privacy protected TensorFlow model."""

from tensorflow_privacy import DPModel

from sacroml.safemodel.safemodel import SafeModel


class SafeTFModel(SafeModel, DPModel):
    """Privacy Protected tensorflow_privacy DP-SGD subclass of Keras model."""

    # pylint:disable=super-init-not-called
    def __init__(
        self,
        l2_norm_clip: float,
        noise_multiplier: float,
        use_xla: bool,
        *args: any,
        **kwargs: any,
    ) -> None:
        """Create model and apply constraints to parameters."""
        raise NotImplementedError



"""Factory for running attacks."""

import logging

import yaml

from sacroml.attacks.attribute_attack import AttributeAttack
from sacroml.attacks.likelihood_attack import LIRAAttack
from sacroml.attacks.structural_attack import StructuralAttack
from sacroml.attacks.target import Target
from sacroml.attacks.worst_case_attack import WorstCaseAttack

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

registry: dict = {
    "attribute": AttributeAttack,
    "lira": LIRAAttack,
    "structural": StructuralAttack,
    "worstcase": WorstCaseAttack,
}


def create_attack(attack_name: str, **kwargs: dict) -> None:
    """Instantiate an attack."""
    if attack_name in registry:
        return registry[attack_name](**kwargs)
    raise ValueError(f"Unknown Attack: {attack_name}")


def attack(target: Target, attack_name: str, **kwargs: dict) -> dict:
    """Create and execute an attack on a target."""
    attack_obj = create_attack(attack_name, **kwargs)
    return attack_obj.attack(target)


def run_attacks(target_dir: str, attack_filename: str) -> None:
    """Run attacks given a target and attack configuration.

    Parameters
    ----------
    target_dir : str
        Name of a directory containing target.yaml.
    attack_filename : str
        Name of a YAML file containing an attack configuration.
    """
    logger.info("Preparing Target")
    target = Target()
    target.load(target_dir)

    logger.info("Preparing Attacks")
    with open(attack_filename, encoding="utf-8") as f:
        config = yaml.safe_load(f)

    logger.info("Running Attacks")

    for attack_cfg in config["attacks"]:
        name = attack_cfg["name"]
        params = attack_cfg["params"]
        attack(target=target, attack_name=name, **params)

    logger.info("Finished running attacks")



"""Prototypes of privacy safe model wrappers."""

from __future__ import annotations

import copy
import datetime
import getpass
import json
import logging
import pathlib
import pickle
from pickle import PicklingError
from typing import Any

import joblib
from dictdiffer import diff

from sacroml.attacks.factory import attack
from sacroml.attacks.target import Target

# pylint : disable=too-many-branches
from .reporting import get_reporting_string

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def check_min(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Check minimum value constraint.

    Parameters
    ----------
    key : string
         The dictionary key to examine.
    val : Any Type
         The expected value of the key.
    cur_val : Any Type
         The current value of the key.
    ..

    Returns
    -------
    msg : string
         A message string.
    disclosive : bool
         A boolean value indicating whether the model is potentially disclosive.
    """
    if isinstance(cur_val, (int, float)):
        if cur_val < val:
            disclosive = True
            msg = get_reporting_string(
                name="less_than_min_value", key=key, cur_val=cur_val, val=val
            )
        else:
            disclosive = False
            msg = ""
        return msg, disclosive

    disclosive = True
    msg = get_reporting_string(
        name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
    )
    return msg, disclosive


def check_max(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Check maximum value constraint.

    Parameters
    ----------
    key : string
         The dictionary key to examine.
    val : Any Type
         The expected value of the key.
    cur_val : Any Type
         The current value of the key.

    Returns
    -------
    msg : string
         A message string.
    disclosive : bool
         A boolean value indicating whether the model is potentially disclosive.
    """
    if isinstance(cur_val, (int, float)):
        if cur_val > val:
            disclosive = True
            msg = get_reporting_string(
                name="greater_than_max_value", key=key, cur_val=cur_val, val=val
            )
        else:
            disclosive = False
            msg = ""
        return msg, disclosive

    disclosive = True
    msg = get_reporting_string(
        name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
    )
    return msg, disclosive


def check_equal(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Check equality value constraint.

    Parameters
    ----------
    key : string
         The dictionary key to examine.
    val : Any Type
         The expected value of the key.
    cur_val : Any Type
         The current value of the key.

    Returns
    -------
    msg : string
         A message string.
    disclosive : bool
         A boolean value indicating whether the model is potentially disclosive.
    """
    if cur_val != val:
        disclosive = True
        msg = get_reporting_string(
            name="different_than_fixed_value", key=key, cur_val=cur_val, val=val
        )
    else:
        disclosive = False
        msg = ""
    return msg, disclosive


def check_type(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Check the type of a value.

    Parameters
    ----------
    key : string
         The dictionary key to examine.
    val : Any Type
         The expected value of the key.
    cur_val : Any Type
         The current value of the key.

    Returns
    -------
    msg : string
         A message string.
    disclosive : bool
         A boolean value indicating whether the model is potentially disclosive.
    """
    if type(cur_val).__name__ != val:
        disclosive = True
        msg = get_reporting_string(
            name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
        )
    else:
        disclosive = False
        msg = ""
    return msg, disclosive


class SafeModel:  # pylint: disable = too-many-instance-attributes
    """Privacy protected model base class.

    Attributes
    ----------
    model_type : string
          A string describing the type of model. Default is "None".
    model:
          The Machine Learning Model.
    saved_model:
          A saved copy of the Machine Learning Model used for comparison.
    ignore_items : list
          A list of items to ignore when comparing the model with the
          saved_model.
    examine_separately_items : list
          A list of items to examine separately. These items are more
          complex datastructures that cannot be compared directly.
    filename : string
          A filename to save the model.
    researcher : string
          The researcher user-id used for logging

    Examples
    --------
    >>> safeRFModel = SafeRandomForestClassifier()
    >>> safeRFModel.fit(X, y)
    >>> safeRFModel.save(name="safe.pkl")
    >>> safeRFModel.preliminary_check()
    >>> safeRFModel.request_release(
    ...     path="safe",
    ...     ext="pkl",
    ...     target=target,
    ... )
    WARNING: model parameters may present a disclosure risk:
    - parameter min_samples_leaf = 1 identified as less than the recommended min value of 5.
    Changed parameter min_samples_leaf = 5.

    Model parameters are within recommended ranges.
    """

    def __init__(self) -> None:
        """Super class constructor, gets researcher name."""
        self.model_type: str = "None"
        self.model = None
        self.saved_model = None
        self.model_load_file: str = "None"
        self.model_save_file: str = "None"
        self.ignore_items: list[str] = []
        self.examine_seperately_items: list[str] = []
        self.basemodel_paramnames = []
        self.filename: str = "None"
        self.researcher: str = "None"
        self.timestamp: str = "None"
        try:
            self.researcher = getpass.getuser()
        except (ImportError, KeyError, OSError):  # pragma: no cover
            self.researcher = "unknown"

    def get_params(self, deep: bool = True) -> dict:
        """Get a dictionary of parameter values restricted to those expected."""
        the_params = {}
        for key, val in self.__dict__.items():
            if key in self.basemodel_paramnames:
                the_params[key] = val
        if deep:
            pass  # not implemented yet
        return the_params

    def save(self, name: str = "undefined") -> None:
        """Write model to file in appropriate format.

        Note this is overloaded in `SafeKerasClassifer`
        to deal with tensorflow specifics.

        Parameters
        ----------
        name : string
             The name of the file to save.

        Notes
        -----
        Optimizer is deliberately excluded to prevent possible restart to
        training and thus possible back door into attacks.
        """
        self.model_save_file = name
        if self.model_save_file == "undefined":
            print("You must input a name with extension to save the model.")
        else:
            thename = self.model_save_file.split(".")
            if len(thename) == 1:
                print("file name must indicate type as a suffix")
            else:
                suffix = self.model_save_file.split(".")[-1]
                # save to pickle
                if suffix == "pkl" and self.model_type != "KerasModel":
                    with open(self.model_save_file, "wb") as file:
                        try:
                            pickle.dump(self, file)
                        except (TypeError, AttributeError, PicklingError) as type_err:
                            print(
                                "saving a .pkl file is unsupported for model type:"
                                f"{self.model_type}."
                                f"Error message was {type_err}"
                            )
                # save to joblib
                elif suffix == "sav" and self.model_type != "KerasModel":
                    try:
                        joblib.dump(self, self.model_save_file)
                    except (TypeError, AttributeError, PicklingError) as type_err:
                        print(
                            "saving as a .sav (joblib) file is not supported "
                            f"for models of type {self.model_type}."
                            f"Error message was {type_err}"
                        )
                else:
                    print(
                        f"{suffix} file suffix currently not supported "
                        f"for models of type {self.model_type}.\n"
                    )

    def __get_constraints(self) -> dict:
        """Get constraints relevant to the model type from the a read-only file."""
        rules: dict = {}
        rule_path = pathlib.Path(__file__).with_name("rules.json")
        with open(rule_path, encoding="utf-8") as json_file:
            parsed = json.load(json_file)
            rules = parsed[self.model_type]
        return rules["rules"]

    def __apply_constraints(
        self, operator: str, key: str, val: Any, cur_val: Any
    ) -> str:
        """Apply a safe rule for a given parameter."""
        if operator == "is_type":
            if (val == "int") and (type(cur_val).__name__ == "float"):
                self.__dict__[key] = int(self.__dict__[key])
                msg = get_reporting_string(name="change_param_type", key=key, val=val)
            elif (val == "float") and (type(cur_val).__name__ == "int"):
                self.__dict__[key] = float(self.__dict__[key])
                msg = get_reporting_string(name="change_param_type", key=key, val=val)
            else:
                msg = get_reporting_string(
                    name="not_implemented_for_change", key=key, cur_val=cur_val, val=val
                )
        else:
            setattr(self, key, val)
            msg = get_reporting_string(name="changed_param_equal", key=key, val=val)
        return msg

    def __check_model_param(
        self, rule: dict, apply_constraints: bool
    ) -> tuple[str, bool]:
        """Check whether a current model parameter violates a safe rule.

        Optionally fixes violations.
        """
        disclosive: bool = False
        msg: str = ""
        operator: str = rule["operator"]
        key: str = rule["keyword"]
        val: Any = rule["value"]
        cur_val: Any = getattr(self, key)
        if operator == "min":
            msg, disclosive = check_min(key, val, cur_val)
        elif operator == "max":
            msg, disclosive = check_max(key, val, cur_val)
        elif operator == "equals":
            msg, disclosive = check_equal(key, val, cur_val)
        elif operator == "is_type":
            msg, disclosive = check_type(key, val, cur_val)
        else:
            msg = get_reporting_string(
                name="unknown_operator", key=key, val=val, cur_val=cur_val
            )
        if apply_constraints and disclosive:
            msg += self.__apply_constraints(operator, key, val, cur_val)
        return msg, disclosive

    def __check_model_param_and(
        self, rule: dict, apply_constraints: bool
    ) -> tuple[str, bool]:
        """Check whether current model parameters violate a logical AND rule.

        Optionally fixes violations.
        """
        disclosive: bool = False
        msg: str = ""
        for arg in rule["subexpr"]:
            temp_msg, temp_disc = self.__check_model_param(arg, apply_constraints)
            msg += temp_msg
            if temp_disc:
                disclosive = True
        return msg, disclosive

    def __check_model_param_or(self, rule: dict) -> tuple[str, bool]:
        """Check whether current model parameters violate a logical OR rule."""
        disclosive: bool = True
        msg: str = ""
        for arg in rule["subexpr"]:
            temp_msg, temp_disc = self.__check_model_param(arg, False)
            msg += temp_msg
            if not temp_disc:
                disclosive = False
        return msg, disclosive

    def preliminary_check(
        self, verbose: bool = True, apply_constraints: bool = False
    ) -> tuple[str, bool]:
        """Check whether current model parameters violate the safe rules.

        Optionally fixes violations.

        Parameters
        ----------
        verbose : bool
             A boolean value to determine increased output level.

        apply_constraints : bool
             A boolean to determine whether identified constraints are
             to be upheld and applied.

        Returns
        -------
        msg : string
           A message string.
        disclosive : bool
           A boolean value indicating whether the model is potentially
           disclosive.
        """
        disclosive: bool = False
        msg: str = ""
        notok_start = get_reporting_string(name="warn_possible_disclosure_risk")
        ok_start = get_reporting_string(name="within_recommended_ranges")
        rules: dict = self.__get_constraints()
        for rule in rules:
            operator = rule["operator"]
            if operator == "and":
                temp_msg, temp_disc = self.__check_model_param_and(
                    rule, apply_constraints
                )
            elif operator == "or":
                temp_msg, temp_disc = self.__check_model_param_or(rule)
            else:
                temp_msg, temp_disc = self.__check_model_param(rule, apply_constraints)
            msg += temp_msg
            if temp_disc:
                disclosive = True

        msg = notok_start + msg if disclosive else ok_start + msg

        if verbose:
            print("Preliminary checks: " + msg)
        return msg, disclosive

    def get_current_and_saved_models(self) -> tuple[dict, dict]:
        """Copy self.__dict__ and split into dicts for current and saved versions."""
        current_model = {}

        attribute_names_as_list = copy.copy(list(self.__dict__.keys()))

        for key in attribute_names_as_list:
            if key not in self.ignore_items:
                try:
                    value = self.__dict__[key]
                    current_model[key] = copy.deepcopy(value)
                except (copy.Error, TypeError) as key_type:
                    logger.warning("%s cannot be copied", key)
                    logger.warning(
                        "...%s error; %s", str(type(key_type)), str(key_type)
                    )

        saved_model = current_model.pop("saved_model", "Absent")

        # return empty dict if necessary
        if (
            saved_model == "Absent"
            or saved_model is None
            or not isinstance(saved_model, dict)
        ):
            saved_model = {}
        else:
            # final check in case fit has been called twice
            _ = saved_model.pop("saved_model", "Absent")

        return current_model, saved_model

    def examine_seperate_items(
        self, curr_vals: dict, saved_vals: dict
    ) -> tuple[str, bool]:
        """Check model-specific items exist in both current and saved copies."""
        msg = ""
        disclosive = False

        for item in self.examine_seperately_items:
            if curr_vals[item] == "Absent" and saved_vals[item] == "Absent":
                disclosive = True
                msg += get_reporting_string(name="both_item_removed", item=item)

            if curr_vals[item] == "Absent" and saved_vals[item] != "Absent":
                msg += get_reporting_string(name="current_item_removed", item=item)
                disclosive = True

            if saved_vals[item] == "Absent" and curr_vals[item] != "Absent":
                disclosive = True
                msg += get_reporting_string(name="saved_item_removed", item=item)

        if not disclosive:  # ok, so can call mode-specific extra checks
            msg, disclosive = self.additional_checks(curr_vals, saved_vals)

        return msg, disclosive

    def posthoc_check(self) -> tuple[str, bool]:
        """Check whether model has been interfered with since fit() was last run."""
        disclosive = False
        msg = ""

        current_model, saved_model = self.get_current_and_saved_models()
        if len(saved_model) == 0:
            msg = get_reporting_string(name="error_not_called_fit")
            msg += get_reporting_string(name="recommend_do_not_release")
            disclosive = True

        else:
            # remove things we don't care about
            for item in self.ignore_items:
                _ = current_model.pop(item, "Absent")
                _ = saved_model.pop(item, "Absent")

            # break out things that need to be handled/examined in more depth
            curr_separate = {}
            saved_separate = {}
            for item in self.examine_seperately_items:
                curr_separate[item] = current_model.pop(item, "Absent")
                saved_separate[item] = saved_model.pop(item, "Absent")

            # comparison on list of "simple" parameters
            match = list(diff(current_model, saved_model, expand=True))
            num_differences = len(match)
            if num_differences > 0:
                disclosive = True
                msg += get_reporting_string(
                    name="basic_params_differ", length=num_differences
                )
                for this_match in match:
                    if this_match[0] == "change":
                        msg += get_reporting_string(
                            name="param_changed_from_to",
                            key=this_match[1],
                            val=this_match[2][1],
                            cur_val=this_match[2][0],
                        )
                    else:
                        msg += f"{this_match}"

            # comparison on model-specific attributes
            extra_msg, extra_disclosive = self.examine_seperate_items(
                curr_separate, saved_separate
            )
            msg += extra_msg
            if extra_disclosive:
                disclosive = True

        return msg, disclosive

    def additional_checks(
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, bool]:
        """Perform additional posthoc checks.

        Placeholder function for additional posthoc checks e.g. keras. This
        version just checks that any lists have the same contents.

        Parameters
        ----------
        curr_separate : dict
        saved_separate : dict

        Returns
        -------
        msg : string
            A message string.
        disclosive : bool
            A boolean value to indicate whether the model is potentially disclosive.

        Notes
        -----
        Posthoc checking makes sure that the two dicts have the same set of
        keys as defined in the list self.examine_separately.
        """
        msg = ""
        disclosive = False
        for item in self.examine_seperately_items:
            if isinstance(curr_separate[item], list):
                if len(curr_separate[item]) != len(saved_separate[item]):
                    msg += (
                        f"Warning: different counts of values for parameter {item}.\n"
                    )
                    disclosive = True
                else:
                    for i in range(len(saved_separate[item])):
                        difference = list(
                            diff(curr_separate[item][i], saved_separate[item][i])
                        )
                        if len(difference) > 0:
                            msg += (
                                f"Warning: at least one non-matching value "
                                f"for parameter list {item}.\n"
                            )
                            disclosive = True
                            break

        return msg, disclosive

    def request_release(self, path: str, ext: str, target: Target = None) -> None:
        """Save model and create a report for the TRE output checkers.

        Parameters
        ----------
        path : string
            Path to save the outputs.
        ext : str
            File extension defining the model saved format, e.g., "pkl" or "sav".
        target : attacks.target.Target
            Contains model and dataset information.

        Notes
        -----
        If target is not null, then worst case MIA and attribute inference
        attacks are called via run_attack.
        """
        # perform checks
        msg_prel, disclosive_prel = self.preliminary_check(verbose=False)
        msg_post, disclosive_post = self.posthoc_check()
        # prepare results
        output: dict = {
            "researcher": self.researcher,
            "model_type": self.model_type,
            "details": msg_prel,
        }
        if hasattr(self, "k_anonymity"):
            output["k_anonymity"] = str(self.k_anonymity)
        if not disclosive_prel and not disclosive_post:
            output["recommendation"] = "Proceed to next step of checking"
        else:
            output["recommendation"] = "Do not allow release"
            output["reason"] = msg_prel + msg_post
        # Run attacks programmatically if possible
        if target is not None:
            for attack_name in ["worstcase", "lira", "attribute"]:
                output[f"{attack_name}_results"] = self.run_attack(
                    target, attack_name, path
                )
        # add timestamp
        now = datetime.datetime.now()
        self.timestamp = str(now.strftime("%Y-%m-%d %H:%M:%S"))
        output["timestamp"] = self.timestamp
        data = [output]
        # save output
        if target is None:
            target = Target(model=self)
        target.add_safemodel_results(data)
        target.save(path, ext)

    def run_attack(
        self,
        target: Target,
        attack_name: str,
        output_dir: str = "outputs_safemodel",
    ) -> dict:
        """Run a specified attack on the trained model and save report to file.

        Parameters
        ----------
        target : Target
            The target in the form of a Target object.
        attack_name : str
            Name of the attack to run.
        output_dir : str
            Name of the directory to store JSON and PDF reports.

        Returns
        -------
        dict
            Metadata results.
        """
        try:
            params = {"output_dir": output_dir}
            output = attack(target=target, attack_name=attack_name, **params)
            metadata = output["metadata"]
        except ValueError:
            metadata = {}
            metadata["outcome"] = "unrecognised attack type requested"
        logger.info("attack %s, metadata %s", attack_name, metadata)
        return metadata

    def __str__(self) -> str:  # pragma: no cover
        """Return string with model description."""
        return self.model_type + " with parameters: " + str(self.__dict__)
